#Requires AutoHotkey v2.0+
#SingleInstance Force

; ----------------------------------------
; define
; ----------------------------------------

; デスクトップ情報
DESKTOP_CLASS := "ahk_class Progman"
DESKTOP_CLASS2 := "ahk_class WorkerW"
DESKTOP_EXE := "ahk_exe explorer.exe"
; タスクトレイのアイコン
BASE_DIR := "C:\tool\keyremap\"
TRAYICON_LIGHT := BASE_DIR . "images\light\"
TRAYICON_DARK := BASE_DIR . "images\dark\"
REG_THEMES_PATH := "HKCU\SOFTWARE\Microsoft\Windows\CurrentVersion\Themes\Personalize"
REG_THEMES_KEY := "AppsUseLightTheme"
; 仮想デスクトップ
VDA_PATH := BASE_DIR . "VirtualDesktopAccessor.dll"
hVirtualDesktopAccessor := DllCall("LoadLibrary", "Str", VDA_PATH, "Ptr")
GetDesktopCountProc := DllCall("GetProcAddress", "Ptr", hVirtualDesktopAccessor, "AStr", "GetDesktopCount", "Ptr")
GoToDesktopNumberProc := DllCall("GetProcAddress", "Ptr", hVirtualDesktopAccessor, "AStr", "GoToDesktopNumber", "Ptr")
GetCurrentDesktopNumberProc := DllCall("GetProcAddress", "Ptr", hVirtualDesktopAccessor, "AStr", "GetCurrentDesktopNumber", "Ptr")
; MoveWindowToDesktopNumberProc := DllCall("GetProcAddress", "Ptr", hVirtualDesktopAccessor, "AStr", "MoveWindowToDesktopNumber", "Ptr")

; グローバル変数
previousInfo := ["", "", ""]
previousRect := [0, 0, 0, 0]
iconNum := 0

; ----------------------------------------
; helper function
; ----------------------------------------

; アクティブ画面のモニター番号取得
getMonitorNumberOfActiveWindow()
{
    WinGetPos(&x, &y, , , "A")
    monitorCount := MonitorGetCount()
    Loop monitorCount {
        MonitorGet(A_Index, &left, &top, &right, &bottom)
        if (x >= left
            and x <= right
            and y >= top
            and y <= bottom) {
            return A_Index
        }
    }
    return 1
}

; ディスプレイ情報取得
getDisplayInfo(monitorNum, &x, &y, &width, &height)
{
    if (monitorNum.Length > 0) {
        displayNum := monitorNum[1]
    } else {
        displayNum := getMonitorNumberOfActiveWindow()
    }
    MonitorGetWorkArea(displayNum, &left, &top, &right, &bottom)
    x := left
    y := top
    width := right - left
    height := bottom - top
}

; 画面サイズMAXの要素位置取得
getMaxValueIndex(array)
{
    index := 0
    value := 0
    Loop array.Length {
        if (array[A_Index] > value) {
            index := A_Index
            value := array[A_Index]
        }
    }
    return index
}

; 再押下ごとの要素位置取得
getRotateIndex(activeHwnd, windowPosition, array, direction*)
{
    previousHwnd := previousInfo[1]
    previousPosition := previousInfo[2]
    previousIndex := previousInfo[3]
    if ((previousHwnd == activeHwnd) and
    (previousPosition == windowPosition)) {
        forward := true
        if (direction.Length > 0) {
            if (GetKeyState("Shift") or (direction[1] == "reverse"))
                forward := false
            if (GetKeyState("Shift") and (direction[1] == "reverse"))
                forward := true
        } else {
            if (GetKeyState("Shift"))
                forward := false
        }
        if (forward) {
            if (previousIndex != array.Length)
                return previousIndex + 1
            else
                return 1
        } else {
            if (previousIndex == 1)
                return array.Length
            else
                return previousIndex - 1
        }
    }
    return 1
}

; 画面位置取得
getWindowPosition(monitorNum, activeHwnd, windowPosition, &x, &y, &width, &height)
{
    global previousInfo
    global previousRect

    ; ディスプレイ情報取得
    getDisplayInfo(monitorNum, &x, &y, &width, &height)

    ; 画面サイズ補正
    ; 画面の左,右,下に隙間が空くので補正する。
    WinGetPos &a, , , , "A"
    WinGetClientPos &b, , , , "A"
    diff := Abs(a - b)
    if (diff == 0)
        edge := 0
    else
        edge := diff - 1

    ; 前回画面位置
    previousPosition := previousInfo[2]
    previousIndex := previousInfo[3]
    previousX := previousRect[1]
    previousWidth := previousRect[3]

    ; 画面位置計算
    if (windowPosition == "1") {
        rotateIndex := getRotateIndex(activeHwnd, windowPosition, LEFT_WIDTH)
        x := x - edge
        y := y
        width := width * LEFT_WIDTH[rotateIndex] + edge * 2
        height := height + edge
    }
    if (windowPosition == "2") {
        rotateIndex := getRotateIndex(activeHwnd, windowPosition, CENTER_WIDTH)
        x := width * (1 - CENTER_WIDTH[rotateIndex]) / 2 - edge
        y := y
        width := width * CENTER_WIDTH[rotateIndex] + edge * 2
        height := height + edge
    }
    if (windowPosition == "3") {
        rotateIndex := getRotateIndex(activeHwnd, windowPosition, RIGHT_WIDTH)
        x := width * (1 - RIGHT_WIDTH[rotateIndex]) - edge
        y := y
        width := width * RIGHT_WIDTH[rotateIndex] + edge * 2
        height := height + edge
    }
    if (windowPosition == "q") {
        rotateIndex := getRotateIndex(activeHwnd, windowPosition, UPPER_HEIGHT)
        x := x - edge
        y := y
        if ((previousPosition == "1") or (previousPosition == "q") or (previousPosition == "a")) {
            width := previousWidth
        } else {
            width := width / 2 + edge * 2
        }
        height := height * LOWER_HEIGHT[rotateIndex] + edge
    }
    if (windowPosition == "a") {
        rotateIndex := getRotateIndex(activeHwnd, windowPosition, LOWER_HEIGHT)
        x := x - edge
        y := height * (1 - LOWER_HEIGHT[rotateIndex]) + y
        if ((previousPosition == "1") or (previousPosition == "q") or (previousPosition == "a")) {
            width := previousWidth
        } else {
            width := width / 2 + edge * 2
        }
        height := height * LOWER_HEIGHT[rotateIndex] + edge
    }
    if (windowPosition == "w") {
        rotateIndex := getRotateIndex(activeHwnd, windowPosition, UPPER_HEIGHT)
        x := width / 2 - edge
        y := y
        if ((previousPosition == "3") or (previousPosition == "w") or (previousPosition == "s")) {
            x := previousX
            width := previousWidth
        } else {
            width := width / 2 + edge * 2
        }
        height := height * UPPER_HEIGHT[rotateIndex] + edge
    }
    if (windowPosition == "s") {
        rotateIndex := getRotateIndex(activeHwnd, windowPosition, LOWER_HEIGHT)
        x := width / 2 - edge
        y := height * (1 - LOWER_HEIGHT[rotateIndex]) + y
        if ((previousPosition == "3") or (previousPosition == "w") or (previousPosition == "s")) {
            x := previousX
            width := previousWidth
        } else {
            width := width / 2 + edge * 2
        }
        height := height * LOWER_HEIGHT[rotateIndex] + edge
    }
    if (windowPosition == "e") {
        rotateIndex := getRotateIndex(activeHwnd, windowPosition, UPPER_HEIGHT)
        x := x - edge
        y := y
        width := width + edge * 2
        height := height * UPPER_HEIGHT[rotateIndex] + edge
    }
    if (windowPosition == "d") {
        rotateIndex := getRotateIndex(activeHwnd, windowPosition, LOWER_HEIGHT)
        x := x - edge
        y := height * (1 - LOWER_HEIGHT[rotateIndex]) + y
        width := width + edge * 2
        height := height * LOWER_HEIGHT[rotateIndex] + edge
    }
    ; 矢印キー（上/下）
    if (windowPosition == "Up") {
        rotateIndex := getRotateIndex(activeHwnd, windowPosition, CENTER_WIDTH)
        x := width * (1 - CENTER_WIDTH[rotateIndex]) / 2 - edge
        y := y
        width := width * CENTER_WIDTH[rotateIndex] + edge * 2
        height := height + edge
    }
    if (windowPosition == "Down") {
        rotateIndex := getRotateIndex(activeHwnd, windowPosition, CENTER_WIDTH, "reverse")
        x := width * (1 - CENTER_WIDTH[rotateIndex]) / 2 - edge
        y := y
        width := width * CENTER_WIDTH[rotateIndex] + edge * 2
        height := height + edge
    }
    ; 矢印キー（左/右）
    previousPosition := previousInfo[2]
    previousIndex := previousInfo[3]
    if (windowPosition == "Left") {
        if ((previousPosition == "Right") and
        (previousIndex != getMaxValueIndex(RIGHT_WIDTH))) {
            windowPosition := "Right"
            rotateIndex := getRotateIndex(activeHwnd, windowPosition, RIGHT_WIDTH)
            x := width * (1 - RIGHT_WIDTH[rotateIndex]) - edge
            y := y
            width := width * RIGHT_WIDTH[rotateIndex] + edge * 2
            height := height + edge
        } else {
            rotateIndex := getRotateIndex(activeHwnd, windowPosition, LEFT_WIDTH, "reverse")
            x := x - edge
            y := y
            width := width * LEFT_WIDTH[rotateIndex] + edge * 2
            height := height + edge
        }
    } else if (windowPosition == "Right") {
        if ((previousPosition == "Left") and
        (previousIndex != getMaxValueIndex(LEFT_WIDTH))) {
            windowPosition := "Left"
            rotateIndex := getRotateIndex(activeHwnd, windowPosition, LEFT_WIDTH)
            x := x - edge
            y := y
            width := width * LEFT_WIDTH[rotateIndex] + edge * 2
            height := height + edge
        } else {
            rotateIndex := getRotateIndex(activeHwnd, windowPosition, RIGHT_WIDTH, "reverse")
            x := width * (1 - RIGHT_WIDTH[rotateIndex]) - edge
            y := y
            width := width * RIGHT_WIDTH[rotateIndex] + edge * 2
            height := height + edge
        }
    }
    OutputDebug("x=" x " y=" y " width=" width " height=" height " windowPosition=" windowPosition " rotateIndex=" rotateIndex)

    ; 前回の移動先保存
    previousInfo[1] := activeHwnd
    previousInfo[2] := windowPosition
    previousInfo[3] := rotateIndex
    previousRect := [x, y, width, height]
}

; 押下キーの付替え
changePressKey(&windowPosition)
{
    previousIndex := previousInfo[3]
    if (previousIndex != 1)
        return
    previousPosition := previousInfo[2]
    if ((previousPosition == "1") and (windowPosition == "Left"))
        previousInfo[2] := "Left"
    if ((previousPosition == "1") and (windowPosition == "Right"))
        previousInfo[2] := "Left"
    if ((previousPosition == "2") and (windowPosition == "Up"))
        previousInfo[2] := "Up"
    if ((previousPosition == "2") and (windowPosition == "Down"))
        previousInfo[2] := "Down"
    if ((previousPosition == "3") and (windowPosition == "Right"))
        previousInfo[2] := "Right"
    if ((previousPosition == "3") and (windowPosition == "Left"))
        previousInfo[2] := "Right"
    if ((previousPosition == "Up") and (windowPosition == "2"))
        previousInfo[2] := "2"
    if ((previousPosition == "Down") and (windowPosition == "2"))
        previousInfo[2] := "2"
    if ((previousPosition == "Left") and (windowPosition == "1"))
        previousInfo[2] := "1"
    if ((previousPosition == "Right") and (windowPosition == "3"))
        previousInfo[2] := "3"
    if ((previousPosition == "q") and (windowPosition == "Right"))
        windowPosition := "w"
    if ((previousPosition == "q") and (windowPosition == "Down"))
        windowPosition := "a"
    if ((previousPosition == "a") and (windowPosition == "Up"))
        windowPosition := "q"
    if ((previousPosition == "a") and (windowPosition == "Right"))
        windowPosition := "s"
    if ((previousPosition == "w") and (windowPosition == "Down"))
        windowPosition := "s"
    if ((previousPosition == "w") and (windowPosition == "Left"))
        windowPosition := "q"
    if ((previousPosition == "s") and (windowPosition == "Up"))
        windowPosition := "w"
    if ((previousPosition == "s") and (windowPosition == "Left"))
        windowPosition := "a"
}

; 画面位置設定
setWindowPosition(windowPosition, monitorNum*)
{
    try {
        ; デスクトップのアイコンが移動してしまうのでガード
        activeHwnd := WinGetID("A")
        GroupAdd("Desktop", DESKTOP_CLASS " " DESKTOP_EXE)
        GroupAdd("Desktop", DESKTOP_CLASS2 " " DESKTOP_EXE)
        desktopHwnd := WinGetID("ahk_group Desktop")
        if (activeHwnd == desktopHwnd)
            return

        ; 画面の最大化を解除
        if (WinGetMinMax(activeHwnd) == 1)
            WinRestore(activeHwnd)

        ; 押下キーの付替え
        changePressKey(&windowPosition)

        ; 移動位置取得
        getWindowPosition(monitorNum, activeHwnd, windowPosition, &x, &y, &width, &height)

        ; 画面移動
        WinMove(x, y, width, height, activeHwnd)
    }
}

; ----------------------------------------
; tasktray
; ----------------------------------------

GetDesktopCount()
{
    count := DllCall(GetDesktopCountProc, "Int")
    return count
}

GoToDesktopNumber(number)
{
    DllCall(GoToDesktopNumberProc, "Int", number - 1)
}

GetCurrentDesktopNumber()
{
    number := DllCall(GetCurrentDesktopNumberProc, "Int")
    number++ ; デスクトップ番号を1オリジンで指定するため + 1
    return number
}

; タスクトレイ更新
updateTaskTray()
{
    global iconNum

    curNum := GetCurrentDesktopNumber()
    if (iconNum != curNum) {
        try {
            ; アイコン更新
            isLightMode := RegRead(REG_THEMES_PATH, REG_THEMES_KEY)
            if (isLightMode)
                TraySetIcon(TRAYICON_LIGHT . curNum . ".png")
            else
                TraySetIcon(TRAYICON_DARK . curNum . ".png")
            ; メニュー更新
            num := 1
            Loop GetDesktopCount() {
                A_TrayMenu.Uncheck("desktop " . num)
                num++
            }
            A_TrayMenu.Check("desktop " . curNum)
            
            iconNum := curNum
        }
    }
}

; タスクトレイ初期化
initTaskTray()
{
    ; DLLが存在する場合：数値アイコン
    if FileExist(VDA_PATH) {
        ; メニュー追加
        A_TrayMenu.Add()
        num := 1
        Loop GetDesktopCount() {
            A_TrayMenu.Add("desktop " . num, MenuHandler)
            num++
        }
        MenuHandler(itemName, itemPos, myMenu) {
            itemCount := DllCall("GetMenuItemCount", "ptr", A_TrayMenu.Handle)
            startNum := itemCount - GetDesktopCount()
            newNum := itemPos - startNum
            GoToDesktopNumber(newNum)
        }
        ; タスクトレイ更新
        updateTaskTray()
        SetTimer(updateTaskTray, 1000)
    } else {
        ; DLLが存在しない場合：Rアイコン
        isLightMode := RegRead(REG_THEMES_PATH, REG_THEMES_KEY)
        if (isLightMode)
            TraySetIcon(TRAYICON_LIGHT . "R.png")
        else
            TraySetIcon(TRAYICON_DARK . "R.png")
    }
}

; タスクトレイ初期化
try initTaskTray()

; ----------------------------------------
; main
;
; リマッピング
; https://ahkscript.github.io/ja/docs/v2/misc/Remap.htm#Remap
; ホットキーモディファイアの記号
; https://ahkscript.github.io/ja/docs/v2/Hotkeys.htm#Symbols
; キーリスト
; https://ahkscript.github.io/ja/docs/v2/KeyList.htm#keyboard
; ----------------------------------------
SendMode "Event"

; 矢印キー × 5
^Up:: sendWrap "{Up 5}"
^Down:: sendWrap "{Down 5}"
^Left:: sendWrap "{Left 5}"
^Right:: sendWrap "{Right 5}"

; 削除
^h:: sendWrap "{BS}"

; 上下左右
^i:: sendWrap "{Up}"
^k:: sendWrap "{Down}"
^j:: sendWrap "{Left}"
^l:: sendWrap "{Right}"

; Home/End
^u:: sendWrap "{Home}"
^o:: sendWrap "{End}"

; 音量調整
^F10:: sendWrap "{VOLUME_MUTE}"
^F11:: sendWrap "{VOLUME_DOWN}"
^F12:: sendWrap "{VOLUME_UP}"

; 画面移動
LEFT_WIDTH   := [0.5, 0.6, 0.7, 0.3, 0.4]
CENTER_WIDTH := [0.5, 0.6, 0.7, 0.3, 0.4]
RIGHT_WIDTH  := [0.5, 0.6, 0.7, 0.3, 0.4]
UPPER_HEIGHT := [0.5, 0.6, 0.7, 0.8, 0.9, 1.0, 0.3, 0.4]
LOWER_HEIGHT := [0.5, 0.6, 0.7, 0.8, 0.9, 1.0, 0.3, 0.4]

; setWindowPositionWrap(windowPosition, monitorNum*)
;   windowPosition：[1|2|3|Left|Up|Down|Right|q|a|w|s|e|d]
;   monitorNum：[1-9]
;   *：省略可能

; 左/中央/右
!*1:: setWindowPositionWrap("1")
!*2:: setWindowPositionWrap("2")
!*3:: setWindowPositionWrap("3")

; 左/中央/右
!*Left:: setWindowPositionWrap("Left")
!*Up:: setWindowPositionWrap("Up")
!*Down:: setWindowPositionWrap("Down")
!*Right:: setWindowPositionWrap("Right")

; 左上/左下/右上/右下
!*q:: setWindowPositionWrap("q")
!*a:: setWindowPositionWrap("a")
!*w:: setWindowPositionWrap("w")
!*s:: setWindowPositionWrap("s")

; 上/下
!*e:: setWindowPositionWrap("e")
!*d:: setWindowPositionWrap("d")


; 不具合対処
; 稀に修飾キーが押下した状態のままになる現象が発生する。
; ホットキー終了時に必ずCtrl/Altキーを上げる。
sendWrap(key)
{
    try
        Send key
    finally
        Send "{Ctrl Up}"
}
setWindowPositionWrap(windowPosition, monitorNum*)
{
    try
        if (monitorNum.Length > 0)
            setWindowPosition(windowPosition, monitorNum[1])
        else
            setWindowPosition(windowPosition)
    finally
        Send "{Alt Up}"
}
^!r:: {
    Send "{Ctrl Down}"
    Send "{Ctrl Up}"
    Send "{Alt Down}"
    Send "{Alt Up}"
    Reload
}


; アクティブ画面の位置サイズを取得・表示する
!F12:: {
    WinGetPos &x, &y, &width, &height, WinGetID("A")
    message := Format('WinMove({1}, {2}, {3}, {4}, WinGetID("A"))', x, y, width, height)
    A_Clipboard := message
    MsgBox(message,, "OK")
}
; 画面位置設定例
!*4:: WinMove(0, 0, 960, 516, WinGetID("A"))
!*5:: WinMove(500, 250, 960, 600, WinGetID("A"))
!*6:: WinMove(960, 516, 960, 516, WinGetID("A"))
